import numpy as np
import networkx as nx
from copy import deepcopy
from nig import NiG
from n import N

class Agent():
    """Network node (agent)"""
    def __init__(self, prior, ind_neighbors, nodes):
        self.prior = prior                            # Node prior
        self.ind_neighbors = ind_neighbors            # Indices of neighbors
        self.nodes = nodes                            # Nodes
        self.nodes_sigma2 = None                      # Nodes noise vars (not always necessary)
        self.dirichlet_params = np.ones(ind_neighbors.size)/ind_neighbors.size # Dirichlet prior for weights
        self.weights_log = [self.weights]             # log of weights
        self.neighbors_posteriors = None              # posterior pdfs of neighbors
        self.neighbors_xt = None                      # neighbors' regressors (not always necessary)
        self.neighbors_yt = None                      # neighbors' observations
#        self.neighbors_sigma2 = None   # property below
        self.yt = None                                # own observation
        self.xt = None                                # own regressor
        self.sigma2 = None                            # own noise var.
        self.t = 0                                    # time counter
        self.unshared_obs = False                     # for internal purposes
    
    @property
    def neighbors_sigma2(self):
        """Neighbors noise vars"""
        return self.nodes_sigma2[self.ind_neighbors]
        
    @property
    def weights(self):
        """Weights for neighbors"""
        return self.dirichlet_params/self.dirichlet_params.sum()
        
    @weights.setter
    def weights(self, w):
        self.dirichlet_params = w

    def weights_update(self):
        """ Update of weights"""
        self._test_have_neighbors_obs()
        logliks = np.zeros(self.ind_neighbors.size)
        for i, posterior in enumerate(self.neighbors_posteriors):
            if self.neighbors_xt is None:
                logliks[i] = posterior.predictive_logpdf(self.neighbors_yt).sum()
            else:
                postpred = [posterior.predictive_logpdf(y, x) 
                            for y,x 
                            in zip(self.neighbors_yt, self.neighbors_xt)]
                logliks[i] = np.sum(postpred)#posterior.predictive_logpdf(self.neighbors_yt,
                             #                        self.neighbors_xt).sum()
        logliks -= logliks.max()
        liks = np.exp(logliks)
        liks /= liks.sum()
        self.dirichlet_params += liks

    def log(self):
        """Logging"""
        self.weights_log.append(self.weights)
        self.prior.log()

    def _test_have_neighbors_obs(self):
        if not self.neighbors_yt or self.unshared_obs:
            if not self.unshared_obs:
                print('Warning: Don\'t have neighbors observations!', flush=True)
                self.unshared_obs = True
            self.neighbors_yt = [self.yt]
            if self.xt is not None:
                self.neighbors_xt = [self.xt]
            return 'own_sigma2'
    
    def adapt(self):     
        """Adaptation step"""
        check = self._test_have_neighbors_obs()
        yt = np.array(self.neighbors_yt)
        if self.neighbors_xt is None:
            if self.nodes_sigma2 is None:
                self.prior.update(yt)            
            else:
                if check == 'own_sigma2':
                    self.prior.update(yt, np.array(self.sigma2))
                else:
                    self.prior.update(yt, self.neighbors_sigma2)
        else:        
            #TODO implement sigma2 as above
            xt = np.array(self.neighbors_xt)
            self.prior.update(yt, xt)
         
    def get_neighbors_posterior(self):
        self.neighbors_posteriors = [deepcopy(self.nodes[i].prior) 
                                     for i in self.ind_neighbors]
                                         
    def get_neighbors_obs(self):
        if self.xt is not None:
            neighbors_xt = [self.nodes[i].xt for i in self.ind_neighbors]
            self.neighbors_xt = np.array(neighbors_xt)
        self.neighbors_yt = [self.nodes[i].yt for i in self.ind_neighbors]
        self.unshared_obs = False
        
    def combine(self):
        """Combination step"""
        xis = [posterior.xi for posterior in self.neighbors_posteriors]
        xis = np.array(xis)
        nus = [posterior.nu for posterior in self.neighbors_posteriors]
        if np.isscalar(xis[0]):
            xi_combined = np.dot(self.weights, xis)
        else:
            xi_combined = (xis * self.weights[:,np.newaxis,np.newaxis]).sum(axis=0)
        nu_combined = np.dot(self.weights, nus)
        self.prior.xi = xi_combined
        self.prior.nu = nu_combined
               
               
class Agent_avg(Agent):
    """Alternative with DMA-style averaging, not in the TSP paper"""
    def __init__(self, prior, ind_neighbors, nodes):
        self.prior = prior
        self.ind_neighbors = ind_neighbors
        self.nodes = nodes
        self.nodes_sigma2 = None
        self.w = np.ones(ind_neighbors.size)/ind_neighbors.size
        self.weights_new = None
        self.weights_log = [self.weights]
        self.neighbors_posteriors = None
        self.yt = None
        self.xt = None
        self.sigma2 = None
        self.combmethod = 'mm'
        self.nocombine = False
        self.forgfactor = .99
        self.stabil_fact = 0.001
        
    def adapt(self):
        #FIXME STABILIZACE VAH
        self.w **= self.forgfactor
        self.w += (self.stabil_fact / self.w.size)
        self.w /= self.w.sum()
        if self.nocombine:
            if self.sigma2 is not None:
                self.prior.update(self.yt, self.xt, self.sigma2)
        else:
            self._weights_update()
            for prior in self.neighbors_posteriors:
                if self.nodes_sigma2 is None:
                    prior.update(self.yt, self.xt.copy())
                else:
                    prior.update(self.yt, self.xt.copy(), self.sigma2)
        
    def combine(self):
        if not self.nocombine:
            if self.combmethod == 'KLdiff':
                super().combine()
            elif self.combmethod == 'mm': #moment matching
                self._moment_matching()
        
    def _moment_matching(self):
        if isinstance(self.neighbors_posteriors[0], N):
            beta_hats = np.array([p.beta_hat.copy() 
                                  for p in self.neighbors_posteriors])
            Vs = np.array([p.V.copy() for p in self.neighbors_posteriors])
            beta_hat_mm = np.dot(self.weights, beta_hats)
            V_mm = 0.
            for i in range(self.weights.size):
                V_mm += self.weights[i] * Vs[i]
                V_mm += self.weights[i] * np.outer(beta_hats[i], beta_hats[i])
            V_mm -= np.outer(beta_hat_mm, beta_hat_mm)
            iV_mm = np.linalg.inv(V_mm)
            self.prior.xi[0,0] = beta_hat_mm.dot(iV_mm).dot(beta_hat_mm)
            self.prior.xi[0,1:] = beta_hat_mm.dot(iV_mm)
            self.prior.xi[1:,0] = beta_hat_mm.dot(iV_mm)
            self.prior.xi[1:,1:] = iV_mm
        
    def _weights_update(self):
        logliks = np.zeros(self.ind_neighbors.size)
        for i, prior in enumerate(self.neighbors_posteriors):
            if self.sigma2 is None:
                logliks[i] = prior.predictive_logpdf(self.yt, self.xt)
            else:
                logliks[i] = prior.predictive_logpdf(self.yt, self.xt, self.sigma2)
        w = np.log(self.weights) + np.array(logliks)
        w -= w.max()
        w = np.exp(w)
        w /= w.sum()
        self.weights_new = w
#        print(w)
        
    def weights_update(self):
        if not self.nocombine:
            self.weights = self.weights_new.copy()
        
    @property 
    def weights(self):
        return self.w
        
    @weights.setter
    def weights(self, w):
        self.w = w
    
            
   
class Network():
    """Network class"""
    
    def __init__(self, adjacency_matrix, prior, agentstype='diff'):
        """Constructor of the network"""
        self.t = 0                                          # time counter
        self.nodes = []                                     # list of nodes
        self.nodes_sigma2 = None                            # nodes' noise variances
        self.adjacency_matrix = adjacency_matrix            # adjacency matrix
        self.degrees = self.adjacency_matrix.sum(axis=0)    # nodes' degrees
        for i, row in enumerate(self.adjacency_matrix):
            ind_neighbors = np.where(row==1)[0]
            if agentstype == 'diff':
                # Diffusion network
                self.nodes.append(Agent(deepcopy(prior), 
                                        ind_neighbors, self.nodes))
            elif agentstype == 'avg':
                self.nodes.append(Agent_avg(deepcopy(prior), 
                                            ind_neighbors, self.nodes))
        for node in self.nodes:
            node.nodes_sigma2 = self.nodes_sigma2
            node.degrees = self.degrees
                                            
    def set_nodes_sigma2(self, list_nodes_sigma2):
        """Set nodes' noise variances"""
        self.nodes_sigma2 = list_nodes_sigma2
        for i, node in enumerate(self.nodes):
            node.nodes_sigma2 = self.nodes_sigma2
            node.sigma2 = self.nodes_sigma2[i]
    
    def set_obs(self, yst, xst=None):
        """Set observations and regressors if present"""
        for i, node in enumerate(self.nodes):
            node.yt = yst[i]
            if xst is not None:
                node.xt = xst[i]
            
    def exchange_obs(self):
        """Exchange observations in the network"""
        for node in self.nodes:
            node.get_neighbors_obs()
            
    def weights_update(self):
        """Update weights"""
        for node in self.nodes:
            node.weights_update()
            
    def adapt(self):
        """Adaptation step"""
        for node in self.nodes:
            node.adapt()
        self.t += 1
        
    def exchange_posteriors(self):
        """Exchange posteriors"""
        for node in self.nodes:
            node.get_neighbors_posterior()
            
    def combine(self):
        """Combination step"""
        for node in self.nodes:
            node.combine()
            
    def log(self):
        """Logging"""
        for node in self.nodes:
            node.log()
            
    def msd_nodes(self, param_name, true_value):
        """Calculate MSDs"""
        nodes_msds = []
        for node in self.nodes:
            node_differences = np.array(node.prior.__dict__[param_name])
            node_differences -= true_value
            node_differences **= 2.
            if len(node_differences.shape) == 1:
                node_rowsums = node_differences
            else:
                node_rowsums = node_differences.sum(axis=1)
            node_msds = np.cumsum(node_rowsums) / np.arange(1, node_rowsums.size+1)
            nodes_msds.append(node_msds)
        
        nodes_msds = np.array(nodes_msds)
        netmsd = nodes_msds.mean(axis=0)
        return (netmsd, nodes_msds)
        
    def net_preds(self):
        from collections import Counter
        net_preds = Counter()
        try:
            for node in self.nodes:
                net_preds.update(node.prior.preds)
            sum_preds = sum(net_preds.values())
            net_preds = dict(net_preds)
            net_preds_rel = net_preds.copy()
            print(net_preds_rel)
            net_preds_rel.update((key, float(val)/float(sum_preds))
                                 for key, val in net_preds_rel.items())
            return (net_preds, net_preds_rel)
        except:
            print('\nNode priors do not have preds attribute!')
            
    def set_static_weights(self, rule):
        #TODO Test with class Agent
        if rule == 'averaging' or rule == 'uniform':
            weights = self.adjacency_matrix / self.adjacency_matrix.sum(axis=1)
            weights = weights.T
        elif rule == 'metropolis':
            degrees_arr = self.adjacency_matrix * self.degrees
            degrees_cmp = degrees_arr.copy().T
            maxima = np.maximum(degrees_arr, degrees_cmp)
            maxima[maxima != 0.] = 1./maxima[maxima != 0]
            np.fill_diagonal(maxima, 0)
            np.fill_diagonal(maxima, 1. - maxima.sum(axis=1))
            weights = maxima
        elif rule == 'laplace-max':
            maxdeg = self.degrees.max()
            weights = self.adjacency_matrix * 1./maxdeg
            np.fill_diagonal(weights, 0)
            np.fill_diagonal(weights, 1. - weights.sum(axis=1))
        elif rule == 'rel-degree':
            degrees_arr = self.adjacency_matrix * self.degrees
            degrees_sums = degrees_arr.sum(axis=1)
            weights = (degrees_arr.T / degrees_sums).T
            print(weights)
        elif rule == 'hastings':
            degrees_arr = self.adjacency_matrix * self.nodes_sigma2
            degrees_arr *= self.degrees
            degrees_cmp = degrees_arr.copy().T
            maxima = np.maximum(degrees_arr, degrees_cmp)
            maxima[maxima != 0.] = 1./maxima[maxima != 0]
            sigma2_arr = (self.adjacency_matrix * self.nodes_sigma2).T
            weights = sigma2_arr * maxima
            np.fill_diagonal(weights, 0)
            np.fill_diagonal(weights, 1. - weights.sum(axis=1))
        for i, node in enumerate(self.nodes):
            node.weights = weights[i][node.ind_neighbors]
            
            
            

#%% Testing garbage code
if __name__ == '__main__':
    np.random.seed(111111)
    from scipy.stats import invwishart as iwishart
    
    nnodes = 5
    beta = np.array([.4, -.8, .3])
    ndat = 200    
    X = np.random.uniform(size=(ndat,nnodes,beta.shape[0]))
    y_clear = X.dot(beta)
    y = y_clear.copy()
    
    noise_sigma2 = np.ones(nnodes) * .2**2
    noise_sigma2[3] = .9**2
#    noise_sigma2[1] = .5**2
   
    for i in range(nnodes):
        y[:,i] += np.random.normal(scale=np.sqrt(noise_sigma2[i]), size=ndat)


#%% Prior Setting
    xi0 = np.diag([.1, .01, .01, .01])
    nu0 = 5.
#    prior = NiG(xi0, nu0)
    prior = N(xi0, nu0)

#%% Network setting
    agentstype = 'avg'
#    adjacency_matrix = np.array([[1, 0, 1], 
#                                 [0, 1, 1],
#                                 [1, 1, 1]])
#    adjacency_matrix = np.eye(3)
    #G = nx.barbell_graph(nnodes//2, 1)
    G = nx.gnp_random_graph(nnodes, 0.2, seed=12345)
    nx.draw_networkx(G)
    adjacency_matrix = nx.adjacency_matrix(G).toarray() + np.eye(nnodes)
#    adjacency_matrix = np.eye(nnodes)
    net = Network(adjacency_matrix, prior, agentstype)
    net.set_nodes_sigma2(noise_sigma2.tolist())
    
#%% Estimation run    
    for node in net.nodes:
        node.nocombine = True
    net.set_static_weights('hastings')
    
    for yst, xst in zip(y, X):
        if agentstype == 'diff':
            net.set_obs(yst, xst)
            net.exchange_obs()
            if net.t > 0:
                net.weights_update()
            net.adapt()
            net.exchange_posteriors()
            net.combine()
            net.log()
        elif agentstype == 'avg':
#            if net.t > 40 and net.t < 60:
#                print(net.t)
#                net.nodes[4].prior.xi = xi0.copy()
#            if net.t > 0:
#                for node in net.nodes:
#                    node.nocombine = False
            net.set_obs(yst, xst)
            net.exchange_posteriors()
            net.adapt()
            net.combine()
            net.weights_update()            
            net.log()
                    
            

#%% Analyses
    MSEs = [np.linalg.norm(node.prior.beta_hat_log - beta) for node in net.nodes]
    print('MMSE: ', np.mean(MSEs))
    print('MSEs: ', MSEs)

#%% Analyses for Barbell graph with 20+1 nodes    
    import matplotlib.pyplot as plt
    plt.figure(2)
    for i in range(nnodes):
        plt.subplot(nnodes, 1, i+1)
        plt.plot(net.nodes[i].weights_log)
        plt.legend(net.nodes[i].ind_neighbors)
        plt.ylabel(str(i))
    
#%% MSES analyses
    mses = net.msd_nodes('beta_hat_log', beta)
    import matplotlib.pyplot as plt
    plt.figure(3)
    plt.plot(mses[0])
    
    for node in net.nodes:
        node.prior.beta_hat_log = np.array(node.prior.beta_hat_log)
    
    plt.figure(4)
    for i in range(beta.size):
        plt.subplot(beta.size,1,i+1)
        plt.hold(True)
        plt.axhline(beta[i])
        for j in range(nnodes):
            plt.plot(net.nodes[j].prior.beta_hat_log[:,i], label=str(j))
        plt.legend()
