import numpy as np
from scipy.stats.distributions import norm

class N():
    """Normal prior"""
    def __init__(self, xi, nu):
        self.xi = xi                            # hyperparameter xi
        self.nu = float(nu)                     # hyperparameter nu
        self.nparams = self.xi.shape[0] - 1.    # number of parameters
        self.beta_hat_log = []                  # log of mean beta_hat
        self.beta_var_log = []                  # log of variance of beta_hat
        
    def update(self, y, X, sigma2):
        """Bayesian update"""
        if len(X.shape) > 1:
            dt = np.hstack((y[:,np.newaxis], X))
            prod = dt.T.dot(dt)
        elif len(X.shape) == 1 and np.isscalar(y):
            dt = np.insert(X, 0, y)
            prod = np.outer(dt, dt)
        self.xi += prod/sigma2
        self.nu += np.size(y)
        
    def log(self):
        """Logging"""
        self.beta_hat_log.append(self.beta_hat)
        self.beta_var_log.append(self.V.diagonal())
        
    @property 
    def beta_hat(self):
        """Estimate of mean value"""
        return np.dot(self.V, self.xi[1:, 0])
        
    @property
    def V(self):
        """Covariance of theta estimates"""
        return np.linalg.inv(self.xi[1:,1:])
        
    @property
    def iV(self):
        """Inverse of V"""
        return self.xi[1:,1:]
        
    def predictive_logpdf(self, y, X, sigma2):
        """Log of predictive pdf"""
        if type(y) is list:
            loc = [x.dot(self.beta_hat) for x in X]
            scales2 = [sig2 + x.dot(self.V).dot(x.T)
                       for x, sig2 in zip(X, sigma2)]
        else:
            loc = X.dot(self.beta_hat)
            scales2 = sigma2 + X.dot(self.V).dot(X.T)
        return norm.logpdf(y, loc=loc, scale=np.sqrt(scales2))
      
      
#----------------------
if __name__ == '__main__':
    xi0 = np.diag([.01, .1, .1, .1])
    nu0 = 3.
    
    sigma2 = .2
    noise_std = np.sqrt(sigma2)
    ndat = 500
    beta = np.array([.5, .1, -1.2])
    X = np.random.uniform(size=(ndat, 3))
    y = X.dot(beta) 
    y += np.random.normal(scale=noise_std, size=y.shape)
    
    prior = N(xi0, nu0)
    for yt, xt in zip(y, X):
        prior.update(yt, xt, noise_std**2)
        print(prior.beta_hat)
        prior.log()
#        print(prior.predictive_logpdf(y[:2], X[:2]))
        
    beta_std_log = np.sqrt(prior.beta_var_log)
    beta_hat_log = np.array(prior.beta_hat_log)
    
    import matplotlib.pylab as plt
    plt.figure(1)
    for i in range(beta.size):
        plt.subplot(beta.size, 1, i+1)
        plt.plot(beta_hat_log[:,i])
        plt.plot(beta_hat_log[:,i]+3.*beta_std_log[:,i], 'k')
        plt.plot(beta_hat_log[:,i]-3.*beta_std_log[:,i], 'k')
        plt.axhline(beta[i], color='red')