import numpy as np
from scipy.special import gammaln, factorial


class PoisProc():
    """Poisson process Po(lambda) estimation class"""
    
    def __init__(self, a, b):
        """Constructor of the gamma prior G(a,b)"""
        self.a = a
        self.b = b
        self.a_log = []
        self.b_log = []
        self.lambda_log = []
        self.mean_log = self.lambda_log
        self.forg = 1.
        
    def update(self, y):
        """Bayesian update by an observation y"""
        self.a *= self.forg
        self.b *= self.forg
        self.a += y.sum()
        self.b += y.size
        
    @property
    def lambd(self):
        """Estimate of lambda"""
        return self.a / self.b
        
    @property
    def xi(self):
        """Hyperparameter xi"""
        return self.a
        
    @xi.setter
    def xi(self, xi):
        self.a = xi

    @property
    def nu(self):
        """Hyperparameter nu"""
        return self.b
        
    @nu.setter
    def nu(self, nu):
        self.b = nu
        
    def predictive_logpdf(self, y):
        result = gammaln(self.a + y) - (self.a + y) * np.log(self.b + 1.)
        result += self.a * np.log(self.b) - gammaln(self.a)
        result -= np.log(factorial(y))
        return result
        
    def predictive_pdf(self, y):
        from scipy.special import gamma, beta
        result = gamma(y) / beta(self.a, y)
        tmp = np.exp(self.a*np.log(self.b) - (self.a + y) * np.log(self.b + 1.))
        result *= tmp
        result /= factorial(y)
        return result
        
    def log(self):
        """Logging"""
        self.mean_log.append(self.a / self.b)
        self.a_log.append(self.a)
        self.b_log.append(self.b)
        

#%% Test
if __name__ == '__main__':
    import matplotlib.pylab as plt
    
    ndat = 500
    lambd = 5.
    np.random.seed(12345)
    y = np.random.poisson(lam=lambd, size=ndat)
    a, b = (1., 1.)
    
    prior = PoisProc(a, b)
    for yt in y:
        prior.update(yt)
        prior.log()
    
    mean_log = np.array(prior.mean_log)
    
    yy = np.array([5., 4, 3, 6.])
    print(np.exp(prior.predictive_logpdf(yy)))
    print(prior.predictive_pdf(yy))
    
    plt.figure(10)
    plt.clf()
    plt.plot(mean_log)
    plt.axhline(lambd)